import java.util.Comparator;





/* Ich bin ein AVL-Baum.
   AVL-Baeume haben eine Wurzel, an dem der gesamte Baum haengt und ein Comparator-Objekt,
   das die Ordnung in dem Baum festlegt.
   Die Anzahl der Knoten wird immer gleich mitgezaehlt, um den Zugriff auf die Groesse des
   Baumes zu beschleunigen. Gleiches gilt fuer die Hoehe (die ja auch fuer das Ausbalancieren
   relevant ist).
   Alle Knoten ausser der Wurzel existieren genau dann, wenn sie einen Wert enthalten. Die
   Wurzel existiert immer. Wenn sie keinen Wert enthaelt, dann ist empty==true.
   Alle angegebenen Laufzeiten sind WorstCase-Laufzeiten. Bei rekursiven Methoden zaehlen
   die aufgerufenen Methoden mit. n bezeichnet dabei immer die Anzahl der Knoten im Baum.
   Achtung! Die Schluesselwerte (also die Eigenschaften der im Baum gespeicherten Objekte,
   die vom gewaehlten Comparator zum Vergleich herangezogen werden) duerfen nicht
   veraendert werden, waehrend die Elemente im Baum gespeichert sind.
 */
public class AVLTree<E> {
    private AVLTreeNode   root;                //die Wurzel eben
    private Comparator<E> comp;               //die Vergleichs-Blackbox
    private int           nodecount = 0;     //die Anzahl der Knoten, leere Wurzeln zaehlen nicht
    private boolean       empty     = true; //Die Wurzel existiert, ist aber leer.



    /* erzeugt einen neuen AVL-Baum. Die Ordnung im Baum wird durch den Comparator bestimmt.
       Laufzeit: O(1)
     */
    public AVLTree(Comparator<E> comp) {
        root      = new AVLTreeNode(); //dieser Baum bekommt schon mal eine Wurzel
        this.comp = comp;             //die Vergleichs-Blackbox speichern
    }



    /* schreibt alle Werte im Baum sortiert in das Array ziel. Das Array muss gross genug sein,
       um alle Werte aufnehmen zu koennen.
       Laufzeit: O(n)
     */
    public void toArray(E[] ziel) {
        toArray(root, ziel, 0);
    }



    /* laedt alle Werte aus dem Array in dem Baum (und verwirft die zuvor im Baum enthaltenen
       Elemente). Das Array wird als Grundlage der Sortierung im Baum verwendet (daher muessen
       die Elemente im Array nach der durch den vom Baum verwendeten Comparator induzierten
       Ordnung sortiert sein. Wenn die Sortierung nicht passt, wird keine Veraenderung des
       Baumes vorgenommen.
       Der Rueckgabewert ist genau dann true, wenn die Werte erfolgreich geladen wurden.
       Laufzeit: O(n)
     */
    public boolean fromArray(E[] quelle) {
        if (quelle.length != 0 && quelle[0] == null)      //ungueltiger Eintrag
            return false;
        for (int i = 1; i < quelle.length; i++)         //Sortierung pruefen ...
            if (quelle[i] == null ||comp.compare(quelle[i - 1], quelle[i]) >= 0)
                return false;                         //... und ggf. abbrechen.
        nodecount = quelle.length;
        root      = new AVLTreeNode();
        fromArray(root, quelle, 0, nodecount - 1); //Werte einlesen.
        return true;
    }



    /* wandelt den Baum in eine einfache ASCII-Graphik um und gibt diese aus.
       Fuer die Ausgabe der Elemente wird <E>.toString() verwendet.
       Laufzeit: O(n*log(n))
     */
    public String toString() {
        if (empty)
            return "(empty)";    //Ein leerer Baum ist nun mal recht strukturlos.
     return toString(root, ""); //Gefuellte Baeume werden gesondert behandelt.
    }



    /* fuegt ein Element in die Baumstruktur ein.
       Ein Element wird nur eingefuegt, wenn der Schluesselwert noch nicht vorkommt.
       Wenn das Einfuegen erfolgreich war, wird true zurueckgegeben, sonst false.
       Laufzeit: O(log(n))
     */
    public boolean insert(E wert) {
        if (wert == null || exists(wert)) //keine null- und nur neue Werte einfuegen
            return false;
        insert(root, wert);             //Wert einfuegen
        nodecount++;                   //Wenn man etwas einfuegt, wird der Baum groesser.
        return true;
    }



    /* gibt genau dann true zurueck, wenn sich das Element bereits im Baum befindet.
       Laufzeit: O(log(n))
     */
    public boolean exists(E wert) {
        if (wert == null)                  //Es gibt keine null-Werte im Baum.
            return false;
        return find(root, wert) != null; //andere Werte kann man suchen.
    }



    /* gibt das Minimum (bzw. null, wenn es nicht existiert) zurueck.
       Laufzeit: O(log(n))
     */
    public E min() {
        return (empty? null: min(root).value);
    }



    /* gibt das Maximum (bzw. null, wenn es nicht existiert) zurueck.
       Laufzeit: O(log(n))
     */
    public E max() {
        return (empty? null: max(root).value);
    }



    /* entfernt das Element value aus dem Baum.
       Es wird genau dann true zurueckgegeben, wenn das Element wirklich existiert hat.
       Laufzeit: O(log(n))
     */
    public boolean remove(E value) {
        if (value == null)     //null-Werte gibt es nicht, also kann man
            return false;     //sie auch nicht entfernen.
        AVLTreeNode node = find(root, value);
        if (node != null) { //Wenn es den Wert wirklich gibt, kann man den zugehoerigen
            remove(node);  //Knoten auch wirklich entfernen
            nodecount--;
            return true;
        }
        return false;
    }



    /* entfernt das Element mit dem Wert value aus dem Baum und gibt es zurueck.
       Laufzeit: O(log(n))
     */
    public E extract(E value) {
        if (value == null) //null-Werte gibt es nicht, daher kann nichts entfernt werden.
            return null;  //null zurueckgeben ist aber kein Problem. ;-)
        AVLTreeNode node = find(root, value);
        if (node == null)       //Wenn es den Wert nicht gibt, kann man das Element auch
            return null;       //nicht Entfernen
        E wert = node.value;
        remove(node);
        nodecount--;
        return wert;
    } 



    /* entfernt die Wurzel aus dem Baum und gibt ihren Wert zurueck.
       Laufzeit: O(1)
     */
    public E extractroot() {
        if (empty)             //Aus leeren Baeumen kann nichts entfernt werden.
            return null;
        E wert = root.value; //Wert lesen
        remove(root);       //Wurzel entfernen
        nodecount--;
        return wert;      //Wert zurueckgeben.
    }



    /* entfernt das minimale Element aus dem Baum und gibt es zurueck.
       Laufzeit: O(log(n))
     */
    public E extractmin() {
        if (empty)
            return null;
        AVLTreeNode node = min(root);
        E wert = node.value;
        remove(node);
        nodecount--;
        return wert;
    } 



    /* entfernt das maximale Element aus dem Baum und gibt es zurueck.
       Laufzeit: O(log(n))
     */
    public E extractmax() {
        if (empty)
            return null;
        AVLTreeNode node = max(root);
        E wert = node.value;
        remove(node);
        nodecount--;
        return wert;
    } 



    /* gibt das Element mit dem Wert value zurueck.
       Laufzeit: O(log(n))
     */
    public E valueOf(E value) {
        if (value == null) //null-Werte gibt es nicht, daher kann nichts entfernt werden.
            return null;  //null zurueckgeben ist aber kein Problem. ;-)
        AVLTreeNode node = find(root, value);
        return (node == null? null : node.value);
    }



    /* gibt die Hoehe des Baumes zurueck (0, wenn der Baum leer ist).
       Laufzeit: O(1)
     */
    public int height() {
        return root.height;
    }



    /* gibt genau dann true zurueck, wenn der Baum leer ist.
       Laufzeit: O(1)
     */
    public boolean isEmpty() {
        return empty;
    }



    /* gibt die Anzahl der im Baum gespeicherten Elemente zurueck.
       Laufzeit: O(1)
     */
    public int size() {
        return nodecount;
    }



    /* sucht in dem Teilbaum, dessen Wurzel node ist, nach dem Knoten mit dem minimalen Wert.
       Laufzeit: O(log(n))
     */
    private AVLTreeNode min(AVLTreeNode node){
        if (node.left == null)   //wenn es keine kleineren Knoten mehr gibt, ist node das Minimum, ...
            return node;
        return min(node.left); //... sonst ist das Mininum weiter unten.
    }



    /* sucht in dem Teilbaum, dessen Wurzel node ist, nach dem Knoten mit dem maximalen Wert.
       Laufzeit: O(log(n))
     */
    private AVLTreeNode max(AVLTreeNode node){
        if (node.right == null)   //wenn es keine groesseren Knoten mehr gibt, ist node das Maximum, ...
            return node;
        return max(node.right); //... sonst ist das Maxinum weiter unten.
    }



    /* fuegt ein Element in den Teilbaum, dessen Wurzel node ist, ein.
       Laufzeit: O(log(n))
     */
    private void insert(AVLTreeNode node, E wert){
        if (node.value == null) {   //node hat Platz fuer das neue Element
            node.value  = wert;    //Wert speichern
            node.height = 1;      //Hoehe ist, da der Baum leer ist und somit keine Kinder existieren
            empty       = false; //Jetzt ist der Baum ganz sicher nicht mehr leer.
            return;             //Fertig.
        }
        if (comp.compare(wert, node.value) > 0) {
            //der Wert muss in den rechten Teilbaum
            if (node.right == null) {                   //Wenn es noch keinen rechten Teilbaum gibt, ...
                node.right        = new AVLTreeNode(); //... dann machen wir schnell einen.
                node.right.parent = node;
            }
            insert(node.right, wert); //Wert speichern
        } else {
            //der Wert muss in den linken Teilbaum
            if (node.left == null) {                   //Wenn es noch keinen linken Teilbaum gibt, ...
                node.left        = new AVLTreeNode(); //... dann pflanzen wir halt einen neuen.
                node.left.parent = node;
            }
            insert(node.left, wert);
        }
        if (!isBalanced(node)) //durch das Erzeugen der neuen Knoten koennte der Baum u.U. nicht
            rotate(node);     //mehr balanciert sein.
    }



    /* sucht nach dem Knoten, der das gewuenschte Element enthaelt.
       Wenn das Element nicht existiert, wird null zurueckgegeben.
       Laufzeit: O(log(n))
     */
    private AVLTreeNode find(AVLTreeNode node, E wert) {
        if (empty)      //Wenn der Baum leer ist, ist das Element nicht drin.
            return null;
        if (comp.compare(wert, node.value) >= 0) {
            //der gesuchte Wert befindet sich nicht im linken Teilbaum
            if (comp.compare(wert, node.value) == 0) //wir haben das Element gefunden
                return node;
            return (node.right == null? null : find(node.right, wert)); //Suche im rechten Teilbaum fortsetzen
        } else
            //das Element befindet sich im linken Teilbaum
            return (node.left  == null? null : find(node.left , wert)); //Suche im linken Teilbaum fortsetzen
    }



    /* Sucht den Knoten mit dem naechstgroesseren Element im Baum.
       Wenn dieser Knoten nicht existiert, wird null zurueckgegeben.
       Laufzeit: O(log(n))
     */
    private AVLTreeNode next(AVLTreeNode node) {
        if (node.right != null)      //Wenn es einen rechten Teilbaum gibt, ist der
            return min(node.right); //Nachfolger das Minimum in diesem Teilbaum
        //Sonst gehen wir so lange nach oben, bis wir einen Nachfolger auftreiben.
        AVLTreeNode p = node.parent;
        while (p != null && comp.compare(p.value, node.value) < 0)
            p = p.parent;
        return p;
    }



    /* Sucht den Knoten mit dem naechstkleineren Element im Baum.
       Wenn dieser Knoten nicht existiert, wird null zuruckgegeben.
       Laufzeit: O(log(n))
     */
    private AVLTreeNode previous(AVLTreeNode node) {
        if (node.left != null)      //Wenn es einen linken Teilbaum gibt, ist der
            return max(node.left); //Vorgaenger das Maximum in diesem Teilbaum
         //Wenn es keinen linken Teilbaum gibt, gehen wir so lange nach oben, bis wir einen
        //Knoten mit einem kleineren Element finden, oder die Wurzel erreichen.
        AVLTreeNode p = node.parent;
        while (p != null && comp.compare(p.value, node.value) > 0)
            p = p.parent;
        return p;
    }



    /* entfernt den Knoten node aus dem Baum.
       Laufzeit: O(log(n))
     */
    private void remove(AVLTreeNode node) {
        if (node.parent == null) {
            //node ist die Wurzel (d.h. node==root)
            if (node.left == null) {
                //node hat keinen linken Teilbaum
                if (node.right != null) {
                    //node hat nur den rechten Teilbaum => Gesamtbaum = rechter Teilbaum
                    root        = node.right;
                    root.parent = null;
                } else {
                    //node hat keine Kinder => Gesamtbaum = leerer Baum
                    empty  = true;
                    root.height = 0;
                }
                return;
            } else if (node.right == null) {
                //node hat nur den linken Teilbaum => Gesamtbaum = linker Teilbaum
                root        = node.left;
                root.parent = null;
                return;
            } else {
                //node hat zwei Kinder
                AVLTreeNode nroot       = next(node);
                AVLTreeNode balanceTest = nroot.parent;
                if (nroot.parent.parent == null) {
                         //node hat als Nachfolger sein rechtes Kind
                        //Dieses wird als neue Wurzel verwendet und erhaelt den linken
                       //Teilbaum von node als neuen linken Teilbaum. (nroot kann
                      //keinen linken Teilbaum haben, weil dort kleinere Werte als in
                     //nroot waeren und diese dann den Nachfolger von node enthalten
                    //wuerden.)
                    root             = nroot;
                    root.parent      = null;
                    root.left        = node.left;
                    root.left.parent = root;
                    if (!isBalanced(root)) //Ggf. muss der Baum noch mal ausbalanciert werden.
                        rotate(root);
                    return;              //Fertig.
                }
                     //node hat als Nachfolger einen Knoten im linken Teilbaum seines rechten Kindes
                    //Dieser Nachfolge wird die neue Wurzel. Der rechte Teilbaum dieses Nachfolgers
                   //wird an seinen ehemaligen Vaterknoten gehaengt. Der linke Teilbaum von node wird
                  //der neue linke Teilbaum der neuen Wurzel nroot. nroot kann keinen linken Teilbaum
                 //haben, weil sich dort kleinere Werte befinden wuerden und nroot damit nicht der
                //Nachfolger von root(==node) waere.
                root              = nroot;
                nroot.parent.left = nroot.right;
                if (nroot.right != null)
                    nroot.right.parent = nroot.parent;
                root.parent       = null;
                root.left         = node.left;
                root.left.parent  = root;
                root.right        = node.right;    
                root.right.parent = root;
                 //Ggf. muessen wir den gesamten Pfad von der Stelle, wo wir den Nachfolger von node
                //aus dem rechten Teilbaum von node entfernt haben, bis zur Wurzel ausbalancieren.
                balanceall(balanceTest);
            }
            return;
        }
        //node hat Eltern
        if (node.left == null) {
             //node hat keinen linken Teilbaum
            //Wir haengen den rechten Teilbaum an die Stelle, wo node an seinem Vorgaenger haengt.
            if (comp.compare(node.value, node.parent.value) > 0)
                //node ist rechtes Kind
                node.parent.right = node.right;
            else
                //node ist linkes Kind
                node.parent.left = node.right;
            if (node.right != null)
                node.right.parent = node.parent;
            balanceall(node.parent);    //Ggf. bis zur Wurzel durchbalancieren
            return;
        } else if (node.right == null) {
             //node hat nur den linken Teilbaum
            //=> Der linke Teilbaum kommt an die Stelle, wo node an seinem Vorgaenger haengt.
            if (comp.compare(node.value, node.parent.value) > 0)
                //node ist rechtes Kind
                node.parent.right = node.left;
            else
                //node ist linkes Kind
                node.parent.left = node.left;
            node.left.parent  = node.parent;
            balanceall(node.parent);    //Ggf. bis zur Wurzel durchbalancieren
            return;
        } else {
            //node hat beide Kinder
            AVLTreeNode nnode       = next(node);
            AVLTreeNode balanceTest = nnode.parent; //Ab hier muessen wir nachher den Baum ausbalancieren
            if (comp.compare(nnode.value, nnode.parent.value) > 0) {
                  //node hat als Nachfolger sein rechtes Kind
                 //Wir gehen analog zu dem Fall vor, wo node die Wurzel war.
                //Hier tragen wir jedoch zusaetzlich nnode als neues Kind des Vorgaengers von node ein.
                if (comp.compare(node.value, node.parent.value) > 0)
                    //node ist rechtes Kind
                    node.parent.right = nnode;
                else
                    //node ist linkes Kind
                    node.parent.left = nnode;
                nnode.parent      = node.parent;
                nnode.left        = node.left;
                node.left.parent  = nnode;
                balanceall(nnode); //der (fast) unvermeidliche Balanceakt.
                return;
            }
                 //node hat als Nachfolger einen Knoten im linken Teilbaum seines rechten Kindes
                //Wir gehen analog zu dem Fall vor, wo node die Wurzel war.
               //Hier tragen wir jedoch zusaetzlich nnode als neues Kind des Vorgaengers von node ein.
               if (comp.compare(node.value, node.parent.value) > 0)
                   //node ist rechtes Kind
                   node.parent.right = nnode;
               else
                   //node ist linkes Kind
                   node.parent.left = nnode;
               nnode.parent.left = nnode.right;
               if (nnode.right != null)
                   nnode.right.parent = nnode.parent;
               nnode.parent      = node.parent;
               nnode.right       = node.right;
               node.right.parent = nnode;
               nnode.left        = node.left;
               node.left.parent  = nnode;
                //Wir haben ganz unten etwas veraendert und muessen nun
               //einmal bis zur Wurzel durchbalancieren
               balanceall(balanceTest);
        }
    }



    /* balanciert den Knoten node aus.
       Dabei koennen der Vorgaenger von node unbalanciert werden, also iterieren wir dies so
       lange, bis wir die Wurzel erreichen.
       Laufzeit: O(log(n))
     */
    private void balanceall(AVLTreeNode node) {
        if (!isBalanced(node))     //Wenn node nicht balanciert ist, dann ausbalancieren
            rotate(node);
        if (node.parent != null) //Weiter mit den Vorgaengern, bis wir die Wurzel erreichen
            balanceall(node.parent);
    }



    /* gibt genau dann true zurueck, wenn der Knoten ausbalanciert ist.
       Nebenbei wird auch die Hoehe des von node ausgehenden Baumes neu berechnet und gespeichert.
       Laufzeit: O(1)
     */
    private boolean isBalanced(AVLTreeNode node) {
        int l = lHeight(node);
        int r = rHeight(node);
        node.height = 1 + Math.max(l, r);
        return Math.abs(l-r)<2;
    }



    /* bestimmt die Hoehe des rechten Teilbaumes.
       Wenn node.right==null gilt, ist die Hoehe 0.
       Laufzeit: O(1)
     */
    private int rHeight(AVLTreeNode node) {
        return (node.right == null? 0 : node.right.height);
    }



    /* bestimmt die Hoehe des linken Teilbaumes.
       Wenn node.left==null gilt, ist die Hoehe 0.
       Laufzeit: O(1)
     */
    private int lHeight(AVLTreeNode node) {
        return (node.left == null? 0 : node.left.height);
    }



    /* balanciert den Baum am Knoten node aus.
       Es wird davon ausgegangen, dass der Baum an dieser Stelle wirklich nicht ausbalanciert
       ist, wenn diese Methode aufgerufen wird.
       Die Einteilung in L-, LR-, R-, RL-Rotationen wurde aus der Vorlesung
       "Theoretische Informatik III" (SS 2007) uebernommen.
       Laufzeit: O(1)
     */
    private void rotate(AVLTreeNode node) {
        AVLTreeNode right  = node.right;
        AVLTreeNode left   = node.left;
        int         rh     = rHeight(node);
        int         lh     = lHeight(node);
        if (node.parent == null) {
            //node ist die Wurzel
            if (rh < lh && lHeight(left) >= rHeight(left)) {
                   //L-Rotation
                  //Das linke Kind (left) von node wird statt diesem als neue Wurzel verwendet.
                 //node wird dann an Stelle des rechten Teilbaumes von left eingefuegt und erhaelt
                //den rechten Teilbaum von left als neuen linken Teilbaum.
                root        = left;
                node.left   = left.right;
                if (left.right != null)
                    left.right.parent = node;
                left.right  = node;
                node.parent = left;
                root.parent = null;
                isBalanced(node);
                isBalanced(left);
            } else if (rh < lh) {
                     //LR-Rotation
                    //z ist die Wurzel des rechten Teilbaumes des linken Kindes (left) von node.
                   //node erhaelt den rechten Teilbaum von z als neuen linken Teilbaum. z wird an der
                  //Stelle von node als neue Wurzel des Gesamtbaumes verwendet. node wird dann zum
                 //rechten Kind von z und left wird zum linken Kind von z. left behaelt seinen linken
                //Teilbaum und bekommt zusaetzlich den rechten Teilbaum von z als neuen rechten Teilbaum.
                AVLTreeNode z = left.right;
                node.left     = z.right;
                if (z.right != null)
                    z.right.parent = node;
                root        = z;
                z.right     = node;
                node.parent = z;
                left.right  = z.left;
                if (z.left != null)
                    z.left.parent = left;
                z.left      = left;
                left.parent = z;
                root.parent = null;
                isBalanced(node);
                isBalanced(left);
                isBalanced(z);
            } else if (rHeight(right) >= lHeight(right)) {
                 //R-Rotation
                //Analog zur L-Rotation mit vertauschtem left und right.
                root         = right;
                right.parent = node.parent;
                node.parent  = right;
                node.right   = right.left;
                if (right.left != null)
                    right.left.parent = node;
                right.left   = node;
                root.parent  = null;
                isBalanced(node);
                isBalanced(right);
            } else {
                 //RL-Rotation
                //Analog zur LR-Rotation mit vertauschtem left und right.
                AVLTreeNode z = right.left;
                node.right    = z.left;
                if (z.left != null)
                    z.left.parent = node;
                root        = z;
                z.left      = node;
                node.parent = z;
                right.left  = z.right;
                if (z.right != null)
                    z.right.parent = right;
                z.right      = right;
                right.parent = z;
                root.parent  = null;
                isBalanced(node);
                isBalanced(right);
                isBalanced(z);
            }
            return;
        }
        //node ist nicht die Wurzel
        boolean rchild = comp.compare(node.value, node.parent.value) > 0;
        if (rh < lh && lHeight(left) >= rHeight(left)) {
               //L-Rotation
              //Das linke Kind (left) von node wird statt diesem als neues Kind des Vorgaengers
             //von node eingefuegt. node wird dann an Stelle des rechten Teilbaumes von left
            //eingefuegt und erhaelt den rechten Teilbaum von left als neuen linken Teilbaum.
            if (rchild)
                node.parent.right = left;
            else
                node.parent.left  = left;
            left.parent = node.parent;
            node.left   = left.right;
            if (left.right != null)
                left.right.parent = node;
            left.right  = node;
            node.parent = left;
            isBalanced(node);
            isBalanced(left);
        } else if (rh < lh) {
                 //LR-Rotation
                //z ist die Wurzel des rechten Teilbaumes des linken Kindes (left) von node.
               //node erhaelt den rechten Teilbaum von z als neuen linken Teilbaum. z wird an der
              //Stelle von node als Kind des Vorgaengers von node verwendet. node wird dann zum
             //rechten Kind von z und left wird zum linken Kind von z. left behaelt seinen linken
            //Teilbaum und bekommt zusaetzlich den rechten Teilbaum von z als neuen rechten Teilbaum.
            AVLTreeNode z = left.right;
            node.left     = z.right;
            if (z.right != null)
                z.right.parent = node;
            if (rchild)
                node.parent.right = z;
            else
                node.parent.left  = z;
            z.parent    = node.parent;
            z.right     = node;
            node.parent = z;
            left.right  = z.left;
            if (z.left != null)
                z.left.parent = left;
            z.left      = left;
            left.parent = z;
            isBalanced(node);
            isBalanced(left);
            isBalanced(z);
        } else if (rHeight(right) >= lHeight(right)) {
             //R-Rotation
            //Analog zur L-Rotation mit vertauschtem left und right.
            if (rchild)
                node.parent.right = right;
            else
                node.parent.left  = right;
            right.parent = node.parent;
            node.parent  = right;
            node.right   = right.left;
            if (right.left != null)
                right.left.parent = node;
            right.left   = node;
            isBalanced(node);
            isBalanced(right);
        } else {
             //RL-Rotation
            //Analog zur LR-Rotation mit vertauschtem left und right.
            AVLTreeNode z = right.left;
            node.right    = z.left;
            if (z.left != null)
                z.left.parent = node;
            if (rchild)
                node.parent.right = z;
            else
                node.parent.left  = z;
            z.parent    = node.parent;
            z.left      = node;
            node.parent = z;
            right.left  = z.right;
            if (z.right != null)
                z.right.parent = right;
            z.right      = right;
            right.parent = z;
            isBalanced(node);
            isBalanced(right);
            isBalanced(z);
        }
    }



    /* erzeugt eine ASCII-Graphik von dem Teilbaum, der node als Wurzel hat.
       shift enthaelt die linken Spalten mit den Informationen ueber den uebergeordneten Teil
       des Baumes
       Laufzeit: O(n*log(n)) (log(n), weil die Graphik immer breiter wird)
     */
    private String toString(AVLTreeNode node, String shift) {
        if (node==null)
            return "";
        String zeile=(shift.length()<2?"":shift.substring(0,shift.length()-2)+" +")+node.value.toString()+"\n";
        return zeile+toString(node.right,shift+" |")+toString(node.left,shift+"  ");
    }



    /* schreibt den Inhalt des Teilbaums, der node als Wurzel hat, sortiert in das Array.
       index gibt das erste zu beschreibende Feld an. Der Rueckgabewert ist der Index des
       letzten beschriebenen Feldes.
       Laufzeit: O(n)
     */
    private int toArray(AVLTreeNode node, E[] ziel, int index) {
        if (node == null || empty) //Wo kein Inhalt ist, muss auch nichts geschrieben werden.
            return index;
        int i = toArray(node.left, ziel, index); //Inhalt des linken Teilbaumes schreiben
        ziel[i] = node.value;                   //eigenen Wert eintragen
        return toArray(node.right, ziel, i+1); //Inhalt des rechten Teilbaums schreiben
       }



    /* laedt den Inhalt des Arrays (genauer: die Felder von start bis ende) in den Baum, der
       node als Wurzel hat. Der Baum wird dabei so aufgebaut, dass er balanciert ist.
       Laufzeit: O(n)
     */
    private void fromArray(AVLTreeNode node, E[] quelle, int start, int ende) {
        if (start > ende)               //Es gibt nichts mehr zum Eintragen.
            return;
        if (empty) {                  //Der erste Eintrag ueberhaupt kommt in die Wurzel.
            root.value  = quelle[start];
            root.height = 1;
            empty       = false;   //Jetzt ist der Baum nicht mehr leer.
            return;
        }
        if (start == ende) {    //Es muss nur noch ein Wert eingetragen werden.
            node.value  = quelle[start];
            node.height = 1;
            return;
        }
        int mitte = (start + ende) / 2;
        node.value  = quelle[mitte];
        if (start == mitte) {               //Es muessen insgesamt zwei Werte eingetragen werden.
            node.right  = new AVLTreeNode(node, quelle[ende]);
            node.height = 2;
            return;
        }
        node.left  = new AVLTreeNode(); //Teilbaeume vorbereiten
        node.right = new AVLTreeNode();
        fromArray(node.left, quelle, start, mitte - 1); //Teilbaeume fuellen.
        fromArray(node.right, quelle, mitte + 1, ende);
        isBalanced(node); //Hoehe dieses Knotens neu berechnen.
    }





    /* Ich bin ein Knoten. */
    private class AVLTreeNode {
        public AVLTreeNode left   = null;   //linkes Kind
        public AVLTreeNode right  = null;  //rechtes Kind
        public AVLTreeNode parent = null; //Papa
        public int         height = 0;   //die Hoehe des Knotens
        public E           value;       //der Inhalt

        /* erzeugt einen neuen Knoten mit parent als Vorgaenger und value als Wert.
           Laufzeit: O(1)
         */
        public AVLTreeNode(AVLTreeNode parent, E value) {
            this.value  = value;
            this.parent = parent;
            height      = 1;
        }

        /* erzeugt einfach nur einen neuen Knoten
           Laufzeit: O(1)
         */
        public AVLTreeNode() {}
    }
}