import java.io.*;
import java.util.*;





/* Ich bin ein Graph.
   Bei der Initialisierung gibt es zwei Moeglichkeiten:
    1. mode==0
     Hier werden zunaechst nur die Knoten hinzugefuegt, die auch eine Kante im Graphen haben.
     Damit erreichen wir einen Aufwand von O(m(log(n)+log(m))). Bei fast leeren Graphen kann
     dies einen kleinen Geschwindigkeitsvorteil gegenueber der zweiten Initialisierungsvariante
     geben.
     Wenn spaeter auf die noch nicht geladenen Knoten zugegriffen wird, werden sie hinzugefuegt.
     Es ist dann auch nachtraeglich moeglich, mit einem Aufwand von O(n) das vollstaendige
     Knoten-Array zu erzeugen.
     Kanten werden sowohl in den Knoten als Adjazenz-Liste (oder vielmehr als Adjazenz-AVL-Baum),
     als auch im Graphen gespeichert. Im Graphen werden die Kanten nach ihrem Gewicht sortiert
     in einem AVL-Baum abgelegt. Kanten mit gleichem Gewicht werden dann nach ihren Knotennummern
     sortiert.
     Das Loeschen eines Knotens aus dem Graphen ist nicht vollstaendig moeglich, da die Anzahl
     der Knoten bereits vorgegeben ist.
     Knoten, die mittels remove geloescht werden, werden im Nachhinein als isolierte Knoten des
     Graphen betrachtet. Ueber das Array vertices kann weiterhin auf sie zugegriffen werden.
     In dem Array createdV kann man sich in O(n) alle Knoten speichern lassen, die in dem AVL-
     Baum abgelegt wurden. Damit koennen (z.B. in fast leeren Graphen) die isolierten Knoten
     uebergangen werden.
     Der Wechsel nach mode==1 ist in O(n+m) moeglich.
    2. mode==1
     Hier werden gleich alle Knoten initialisiert. Der Aufwand ist daher O(n+m), was noetig ist,
     wenn der Gesamtaufwand linear sein soll.
     Die Kanten werden weder im Graphen noch in den Knoten in AVL-Baeumen, sondern in Listen
     gespeichert. Eine Sortierung findet nicht statt.
     Es ist jedoch wieder nachtraeglich moeglich, die Knoten und Kanten in den Baeumen eintragen
     zu lassen. Der Aufwand dafuer ist dann jedoch O(nlog n+mlog m) (=> Moduswechsel nach mode==0).
   Bei allen Laufzeitangaben ist n die Anzahl der Knoten und m die Anzahl der Kanten im Graphen.
   Es wird jeweils die WorstCase-Laufzeit angegeben.
 */
class Graph{
    public  Vertice[]        vertices       = new Vertice[0];  //Knoten-Array
    public  Vertice[]        createdV       = new Vertice[0]; //Array der erzeugten Knoten
    public  boolean          createdVOK     = true;          //createdV initialisiert?
    public  boolean          arrayOK        = true;         //Ist das Knoten-Array erzeugt worden?
    public  int              verticeCount   = 0,           //Anzahl der Knoten (Knoten ohne Kanten werden zunaechst nicht gespeichert)
                             edgeCount      = 0;          //Anzahl der Kanten

    private int              mode           = 0;        //Speichermodus
    private AVLTree<Vertice> verticetree    = null;    //so findet man die Knoten schneller
    private AVLTree<Edge>    edgetree       = null;   //Kantensuche
    private List<Edge>       edgelist       = null;  //unsortierte Kantenliste



    /* liest den Graphen aus einer Datei ein.
       1. mode==0:
        Es werden zunaechst nur die Knoten gespeichert, die auch Kanten haben. Die uebrigen
        werden erst einmal weggelassen.
        Laufzeit: O(m(log(n)+log(m)))
       2. mode==1:
        Es werden alle Knoten initialisiert. Daher wird der Aufwand auch mehr von der Anzahl
        der Knoten beeinflusst als im Fall mode==0.
        Laufzeit: O(n+m)
     */
    public void readfromfile(String datei, int mode) {
        //erst mal alles aufraeumen
        String         zeile  = "";
        BufferedReader fin;
        verticetree = null;
        edgetree    = null;
        edgelist    = null;
        this.mode   = mode;
        arrayOK     = false;
        createdVOK  = false;
        vertices    = null;
        //Daten einlesen
        try{
            fin=new BufferedReader(new FileReader(datei));
            int zl=0;
            //solange noch was da ist...
            while ((zeile = fin.readLine()) != null) {
                String[] zeileninhalt=removeEmptyStrings(zeile.split(" "));
                if (zeileninhalt.length != 0 && !zeileninhalt[0].equals("c")) {
                    if (zeileninhalt[0].equals("p")) {
                        verticeCount = Integer.parseInt(zeileninhalt[1]);  //Knotenzahl laden
                        edgeCount = Integer.parseInt(zeileninhalt[2]);    //Kantenzahl laden
                        if (mode == 1) {                                 //Im Fall mode==1
                            edgelist = new List<Edge>();                //mode==1 => Liste erzeugen
                            vertices = new Vertice[verticeCount];      //wird das gesamten Knoten-Array sofort
                            for (int i = 0; i < verticeCount; i++)    //initialisiert und alle Knoten erzeugt.
                                vertices[i] = new Vertice(i + 1, 1); //Die Knoten sollen nur Listen verwenden.
                            arrayOK = true;                         //Wir haben das Array gerade erzeugt.
                        } else {                                   //mode==0 => AVL-Baeume erzeugen.
                            verticetree = new AVLTree<Vertice>(new VerticeComparator());
                            edgetree    = new AVLTree<Edge>(new EdgeComparator());
                        }
                    } else if (zeileninhalt[0].equals("e"))
                        addEdge(Integer.parseInt(zeileninhalt[1]),
                                Integer.parseInt(zeileninhalt[2]),
                                Integer.parseInt(zeileninhalt[3]));
                }
            }
        }
        catch (FileNotFoundException x) {
            System.out.println(datei+": Diese Datei existiert nicht oder kann nicht gelesen werden."); //Kann ja mal passieren, oder?
            System.exit(-1);
        }
        catch (IOException x) {
            x.printStackTrace();
            System.exit(-1);
        }
        if (mode == 1)                              //Wir hatten eh schon O(n+m) Aufwand, also haben
            for (int i = 0; i < verticeCount; i++) //wir hierfuer auch noch Zeit: Die Adjazenzliste 
                vertices[i].createArray();        //jedes Knoten zu einem Adjazenzarray machen
    }



    /* gibt den Modus zurueck.
       Laufzeit: O(1)
     */
    public int getMode() {
        return mode;
    }



    /* wechselt den Modus.
       1. mode==0 => mode==0
        Laufzeit: O(1)
       2. mode==0 => mode==1
        Laufzeit: O(n+m)
       3. mode==1 => mode==0
        Laufzeit: O(n*log(n)+m*log(m))
       4. mode==1 => mode==1
        Laufzeit: O(1)
     */
    public void switchMode(int mode) {
        if (mode == this.mode)                          //Der Modus stimmt bereits.
            return;
        if (mode == 0) {                              //Wechsel von mode==0 nach mode==1
            edgelist = new List<Edge>();             //Kantenliste erzeugen
            mode     = 1;                           //Modus wechseln
            for (int i = 0; i < verticeCount; i++) //alle Knoten neu erzeugen
                vertices[i] = new Vertice(i + 1, 1);
            Edge[] tempEdges = new Edge[edgetree.size()]; //bisherige Kanten aus
            edgetree.toArray(tempEdges);                 //dem Baum holen
            for (int i = 0; i < tempEdges.length; i++)  //entsprechende Kanten fuer
                addEdge(tempEdges[i].fromVertice.id,   //die neuen Knoten erzeugen und eintragen.
                        tempEdges[i].toVertice.id,
                        tempEdges[i].weight);
            edgetree    = null;
            verticetree = null;
            arrayOK     = true;
            createdVOK  = false;
        } else {               //Wechsel von mode==1 nach mode==0
            verticetree = new AVLTree<Vertice>(new VerticeComparator()); //AVL-Baeume erzeugen
            edgetree    = new AVLTree<Edge>(new EdgeComparator());
            mode        = 0;
            for (int i = 1; i <= verticeCount; i++)                   //neue Knoten erzeugen
                addVertice(i);
            Edge[] tempEdges = new Edge[edgelist.count()];          //Kanten aus der Liste holen
            edgelist.toArray(tempEdges);
            for (int i = 0; i < tempEdges.length; i++)            //neue Kanten einfuegen
                addEdge(tempEdges[i].fromVertice.id,
                        tempEdges[i].toVertice.id,
                        tempEdges[i].weight);
            edgelist = null;                                  //Alles alte ist jetzt schlecht.
            createArray();                                   //Arrays erzeugen
        }
    }



    /* fuegt einen Knoten hinzu.
       1. mode==0
        Es werden nur Knoten mit Nummern von 1 bis verticeCount eingefuegt.
        Laufzeit: O(log(n))
       2. mode==1
        Im Fall mode==1 geschieht hier gar nichts.
        Laufzeit: O(1)
     */
    public void addVertice(int v) {
        if (mode == 1)                          //Im Fall mode==1 wird keine Baumstruktur erzeugt, wo
            return;                            //Knoten eingefuegt werden koennten.
        if (v < 1 || v > verticeCount)        //ungueltige Knotennummer
            return;
        verticetree.insert(new Vertice(v)); //Knoten einfuegen.
        createdVOK = false;
    }



    /* entfernt im Fall von mode==0 einen Knoten und alle im Graphen gespeicherten Kanten,
       die diesen Knoten enthalten.
       1. mode==0
        Laufzeit: O(log(n)+log(m)^2)
       2. mode==1
        Der Knoten wird nicht entfernt.
        Laufzeit: O(1)
     */
    public void removeVertice(int v) {
        if (mode == 1)                                           //falscher Modus
            return;
        Vertice tmp = verticetree.extract(new Vertice(v));     //Erst mal nachsehen, ob es diesen
        if (tmp == null)                                      //Knoten ueberhaubt gibt.
            return;
        if (!tmp.arrayOK)                                   //Wurde das Adjazenzarray des Knotens schon erzeugt?
            tmp.createArray();                             //Wenn nein, neu berechnen.
        for (int i=0; i<tmp.neighbours.length; i++)       //Allen Nachbarn mitteilen, dass jetzt die
            tmp.neighbours[i].toVertice.removeEdge(tmp); //Kante zu diesem Knoten fehlt.
        createdVOK = false;                             //Wir haben einen Knoten aus dem Baum entfernt.
    }



    /* fuegt eine Kante mit dem Gewicht 0 hinzu.
       Es werden nur Kanten eingefuegt, deren Knotennummern im Bereich 1 bis verticeCount
       liegen.
       1. mode==0
        Laufzeit: O(log(n)+log(m))
       2. mode==1
        Laufzeit: O(1)
     */
    public void addEdge(int v1, int v2) {
        addEdge(v1, v2, 0);
    }



    /* fuegt eine Kante zwischen den Knoten mit den Nummern v1 und v2 und mit dem Gewicht weight
       hinzu. Es werden nur Kanten eingefuegt, deren Knotennummern im Bereich 1 bis verticCount
       liegen. Weiterhin muss v1!=v2 gelten.
       1. mode==0
        Laufzeit: O(log(n)+log(m))
       2. mode==1
        Laufzeit: O(1)
     */
    public void addEdge(int v1, int v2, int weight) {
        if (v1 < 1 || v1 > verticeCount || v2 < 1 || v2 > verticeCount) //ungueltige Knotennummern
            return;
        if (mode == 0) {                                        //AVL-Baeume verwenden
            Vertice vert1 = new Vertice(v1);                   //Objekte zur Abfrage
            Vertice vert2 = new Vertice(v2);
            if (verticetree.insert(vert1))                   //Wenn es den Knoten noch nicht gibt, dann
                createdVOK = false;                         //einfuegen und das Array der im Baum be-
            else                                           //findlichen Knoten als falsch markieren,
                vert1 = verticetree.valueOf(vert1);       //sonst den Wert abfragen.
            if (verticetree.insert(vert2))               //Wenn es v2 noch nicht gibt, dann einfuegen,
                createdVOK = false;                     //und das Array der im Baum befindlichen 
            else                                       //Knoten als falsch markieren,
                vert2 = verticetree.valueOf(vert2);   //sonst den Wert abfragen.
            Edge e = new Edge(vert1, vert2, weight); //Kante erzeugen
            vert1.addEdge(e);                       //Kante bei den Knoten eintragen
            vert2.addEdge(e);
            edgetree.insert(e);                   //Kante abspeichern
        } else {                                 //Listen verwenden
            Edge e = new Edge(vertices[v1 - 1], vertices[v2 - 1], weight);
            vertices[v1 - 1].addEdge(e);       //Kante in den Knoten eintragen
            vertices[v2 - 1].addEdge(e);
            edgelist.add(e);                 //Kante in die Kantenliste des Graphen eintragen
        }
    }



    /* entfernt die Kante mit den Knotennummern v1 und v2.
       1. mode==0
        Laufzeit: O(log(n)+log(m))
       2. mode==1
        Kante wird nicht entfernt.
        Laufzeit: O(1)
     */
    public void removeEdge(int v1, int v2) {
        if (mode == 1)                                             //falscher Modus.
            return;
        if (v1 < 1 || v1 > verticeCount || v2 < 1 || v2 > verticeCount) //ungueltige Knotennummern
            return;
        Vertice vert1 = verticetree.valueOf(new Vertice(v1));  //Knotenwerte abfragen
        Vertice vert2 = verticetree.valueOf(new Vertice(v2)); 
        if (vert1 == null || vert2 == null)                  //Wenn einer der Knoten nicht exisitert,
            return;                                         //existiert auch keine Kante dazwischen.
        edgetree.remove(vert1.getEdge(vert2));             //Kante entfernen.
        vert1.removeEdge(vert2);                          //Kante aus den Knoten loeschen.
        vert2.removeEdge(vert1);
    }



    /* gibt den Knoten (Vertice-Objekt) mit der Nummer v zurueck.
       Wenn v nicht im Bereich von 1 bis verticeCount liegt, wird null zurueckgegeben.
       Wenn der Knoten noch nicht im Graphen war (weil er bei der Initialisierung uebergangen
       wurde, da er keine Kanten hat), wird er jetzt eingefuegt.
       1. mode==0
        Laufzeit: O(log(n)), wenn das Array noch nicht bereit ist.
                  O(1)     , wenn das Array bereits berechnet wurde.
       2. mode==1
        Laufzeit: O(1)
     */
    public Vertice vertice(int v) {
        if (v < 1 || v > verticeCount) //ungueltige Knotennummer
            return null;
        if (arrayOK || mode == 1)    //Wenn das Knotenarray vorbereitet ist, geht's schnell.
            return vertices[v - 1];
        Vertice vert = verticetree.valueOf(new Vertice(v)); //Wenn der Knoten existiert, dann Rueckgabe, ...
        if (vert == null) {                                //... sonst neu einfuegen
            verticetree.insert(vert = new Vertice(v));
            createdVOK = false;                          //Der Baum wurde veraendert, also stimmt
        }                                               //das Array nicht mehr.
        return vert;
    }



    /* gibt die Kante zwischen den Knoten v1 und v2 zurueck.
       (Kanten sind in der hiesigen Darstellung immer gerichtet. Eine ungerichtete Kante
        zwischen v1 und v2 wird in jedem Knoten einmal gespeichert.
        Weiterhin wird eine der beiden Richtungen im Kanten-AVL-Baum abgelegt, wenn mode==0,
        sonst wird sie in der Kantenliste gespeichert.)
       1. mode==0
        Laufzeit: O(log(n)+log(m))
       2. mode==1
        Laufzeit: O(m)
         Genauer: O(Anzahl der Nachbarn von vertices[v1-1])
    */
    public Edge edge(int v1, int v2) {
        if (v1 < 1 || v1 > verticeCount || v2 < 1 || v2 > verticeCount) //ungueltige Knotennummern
            return null;
        if (mode == 0) {                                               //AVL-Baeume verwenden
            Vertice vert1 = verticetree.valueOf(new Vertice(v1));     //Knoten abfragen
            Vertice vert2 = verticetree.valueOf(new Vertice(v2));
            if (vert1 == null || vert2 == null)                     //Wenn einer der Knoten nicht
                return null;                                       //existiert, gibt es auch keine
                                                                  //Kante dazwischen.
            return vert1.getEdge(vert2);                         //Kante zurueckgeben.
        } else                                                  //AVL-Baeume nicht verwenden
            return vertices[v1 - 1].getEdge(vertices[v2 - 1]); //Kante zurueckgeben
    }



    /* erzeugt das Knoten-Array.
       Diese Methode ist nur im Modus 0 sinnvoll, da es ja im anderen Fall keine AVL-Baeume zu
       konvertieren gibt.
       1. mode==0
        Laufzeit: O(n)
       2. mode==1
        Hier geschieht einfach nichts.
        Laufzeit: O(1)
     */
     public void createArray() {
         if (mode==1)                                       //falscher Modus.
             return;
         Vertice[] tmp = new Vertice[verticetree.size()]; //Knoten sortiert aus dem Baum holen
         verticetree.toArray(tmp);
         vertices = new Vertice[verticeCount];          //Wir bauen gleich beide Arrays auf
         createdV = new Vertice[verticeCount];         //soviel Mehraufwand ist's ja nicht.
         int k = 0;
         for (int i = 1; i <= verticeCount; i++) {   //neue Knoten einfuegen
             if (tmp[k].id > i)
                 vertices[i - 1] = new Vertice(i); //Die id k ist noch nicht vergeben.
             else
                 vertices[i - 1] = tmp[k];       //Die id k ist schon vergeben.
             createdV[i - 1] = vertices[i - 1];
         }
         verticetree.fromArray(vertices);     //Baum neu aufbauen, damit die neuen Knoten enthalten sind.
         arrayOK    = true;                  //Arrays sind jetzt gueltig.
         createdVOK = true;
     }



    /* erzeugt das Array createdV, d.h. es werden alle Knoten des Knoten-AVL-Baumes in dem Array
       gespeichert. Diese Methode kann man verwenden, wenn man alle Knoten betrachten will, die
       mindestens eine Kante haben, und bisher noch keine weiteren Knoten referenziert hat.
       Diese Methode arbeitet nur im Fall mode==0.
       1. mode==0
        Laufzeit: O(n)
       2. mode==1
        Hier geschieht einfach nichts.
        Laufzeit: O(1)
     */
    public void createCreatedVerticesArray() {
        if (mode == 1)   //Array wurde bereits initialisiert oder falscher Modus.
            return;
        createdV   = new Vertice[verticetree.size()]; //Array erzeugen
        verticetree.toArray(createdV);               //Knoten darin speichern
        createdVOK = true;                          //Array ist in Ordnung.
    }



    /* entfernt leere Strings, die nach dem Splitten der Eingabezeile entstanden sind, aus
       dem String-Array.
       Laufzeit: O(args.length)
        Wenn die Eingabedatei korrekt ist und zwischen je zwei Eintraegen genau ein Leerzeichen
        steht, ist die Laufzeit O(1).
     */
    private static String[] removeEmptyStrings(String[] args) {
        int n=0;
        for (int i = 0; i < args.length; i++)      //Wie viele nichtleere Strings gibt es?
            if (args[i].length() != 0)
                n++;
        String[] ausgabe = new String[n];       //ausreichend grosses Array erzeugen
        n=0;
        for (int i = 0; i < args.length; i++) //Strings im Array speichern
            if (args[i].length() != 0)
                ausgabe[n++] = args[i];
        return ausgabe;
    }





    /* Ich bin eine Vergleichs-Black-Box fuer Knoten.
       Knoten werden durch ihre id (1 bis verticeCount) unterschieden.
       Alle Laufzeiten (bzw. die Laufzeit des Vergleichs) sind O(1).
     */
    private class VerticeComparator implements Comparator<Vertice> {



        /* vergleicht zwei Knoten
         */
        public int compare(Vertice gv1, Vertice gv2) {
            return (int) Math.signum(gv1.id - gv2.id);          //Sortierung nach id.
        }
    }





    /* Ich bin eine Vergleichs-Black-Box fuer Kanten.
       Kanten werden
        1. durch ihr Gewicht
        2. durch das Maximum ihrer Knoten-IDs (bei Gleichheit in 1.)
        3. durch das Minimum ihrer Knoten-IDs (bei Gleichheit in 1. und 2.)
       unterschieden.
       (Warum nun gerade Maximum und Minimum der IDs?
        Nun ja ... damit wird erreicht, dass {a, b} die gleiche Kante wie {b, a} ist, aber
        alle Kanten, die normalerweise unterschiedlich sind, auch unterschiedlich bleiben.)
       Mit dieser Klasse wird im AVL-Baum eine Sortierung nach dem Gewicht erreicht.
       Alle Laufzeiten (bzw. die Laufzeit des Vergleichs) sind O(1).
     */
    private class EdgeComparator implements Comparator<Edge> {



        /* vergleicht zwei Kanten
             Laufzeit: O(1)
         */
        public int compare (Edge e1, Edge e2) {
            int i = (int) Math.signum(e1.weight - e2.weight); //Sortierung nach Gewicht, ...
            if (i == 0)
                i = (int) Math.signum( Math.max(e1.fromVertice.id, e1.toVertice.id) // ... Maximum und ...
                                      -Math.max(e2.fromVertice.id, e2.toVertice.id));
            return (i != 0? i : (int) Math.signum(        //... Minimum der Knoten-IDs.
                       Math.min(e1.toVertice.id, e1.fromVertice.id)
                     - Math.min(e2.toVertice.id, e2.fromVertice.id)));
        }
    }
}