import java.io.*;
import java.util.*;





/* Ich bin ein Knoten.
   Jeder Knoten hat ein paar Eigenschaften visited (=besucht), id (1 bis Graph.verticeCount),
   tag (koennte man vll. mal gebrauchen) und color (=Farbe fuer Faerbungsprobleme).
   Es gibt (wie beim Graphen) zwei Modi:
    1. mode==0
     Der Knoten speichert alle Kanten zu Nachbarn in zwei AVL-Baeumen:
      1. neighbourtreeW:
       Hier werden die Kanten erst nach Gewicht und dann (bei Gleichheit des Gewichts) nach
       der id des anderen Endknotens sortiert.
      2. neighbourtreeID
       Hier werden die Kanten gleich nach der ID des anderen Endknotens sortiert.
       Da wir keine Multigraphen betrachten, reicht das auch aus.
     Wenn man alle Kanten des Knoten schnell durcharbeiten moechte, kann man das neighbours-
     Array verwenden. Dies muss jedoch ggf. erst in O(n) initialisiert werden.
     Das Umschalten in mode==1 ist in O(n) moeglich.
    2. mode==1
     Hier werden die Kanten zu den Nachbarn in einer Liste gespeichert.
     Wenn man darin nach einem bestimmten Nachbarn sucht, ist dies nur in O(n) moeglich.
     Das Umschalten in mode==0 ist in O(n*log(n)) moeglich.
   n ist bei allen Aufwandsbetrachtungen die Anzahl der Nachbarn.
 */
class Vertice{
    public  Edge[]        neighbours      = new Edge[0]; //zum schnellen Abarbeiten aller
                                                        //Kanten, nach Gewicht sortiert, wenn
                                                       //mode==0, sonst unsortiert
    public  boolean       arrayOK         = true;     //Ist das Array schon initialisiert?
    public  boolean       visited         = false;   //wurde der Knoten mal besucht?
    public  int           color           = 0;      //Farbe des Knotens (0 fuer ungefaerbt)
    public  Object        tag             = null;  //Alles, was man sonst noch moechte kommt hier rein.
    public  int           id              = 0;    //Das Erkennungszeichen des Knotens.

    private int           mode            = 0;  //AVL oder nicht AVL - das ist hier die Frage
    private AVLTree<Edge> neighbourtreeW  = new AVLTree<Edge>(new EdgeWeightComparator()); //Hier geht's schnell zu einem Nachbarn.
    private AVLTree<Edge> neighbourtreeID = new AVLTree<Edge>(new EdgeIDComparator());    //Hier geht's schnell zu einem bestimmten Nachbarn.
    private List<Edge>    neighbourlist   = new List<Edge>(); //Liste der Nachbarn zum schnellen Einfuegen.



    /* erzeugt aus der id einen neuen Knoten im Modus 0.
       Laufzeit: O(1)
     */
    public Vertice(int id) {
        this(id, 0);
    }



    /* erzeugt einen neuen Knoten im Modus mode mit der gegebenen id.
       Laufzeit: O(1)
     */
    public Vertice(int id, int mode) {
        this.id   = id;
        this.mode = mode;
    }



    /* gibt den aktuellen Modus zurueck.
       Laufzeit: O(1)
     */
    public int sendMode() {
        return mode;
    }



    /* fuegt eine neue Kante zum Knoten v mit Gewicht 0 hinzu.
       1. mode==0
        Laufzeit: O(log(n))
       2. mode==1
        Laufzeit: O(1)
     */
    public void addEdge(Vertice v) {
        addEdge(v, 0);    //Weiterdelegation an addEdge(Vertice, int)
    }



    /* fuegt eine neue Kante zum Knoten v mit Gewicht weight hinzu.
       1. mode==0
        Laufzeit: O(log(n))
       2. mode==1
        Laufzeit: O(1)
     */
    public void addEdge(Vertice v, int weight) {
        addEdge(new Edge(this, v, weight));    //Weiterdelegation an addEdge(Edge)
    }



    /* fuegt die Kante e ein.
       1. mode==0
        Laufzeit: O(log(n))
       2. mode==1
        Laufzeit: O(1)
     */
    public void addEdge(Edge e) {
        if (mode==0) {
            if (neighbourtreeW.insert(e)) //Im gewichtssortierten Baum speichern, ...
                arrayOK = false;         //... ggf. das Array als ungueltig deklarieren ...
            neighbourtreeID.insert(e);  //... und im id-sortieren Baum ablegen.
        } else {
            neighbourlist.add(e);     //Bei einer Liste geht das Einfuegen schnell,
            arrayOK = false;         //leider stimmt dann auch hier das Array nicht mehr.
        }
    }



    /* entfernt die Kante zum Knoten v, wenn mode==0.
       1. mode==0
        Laufzeit: O(log(n))
       2. mode==1
        Die Kante wird nicht entfernt.
        Laufzeit: O(1)
     */
    public void removeEdge(Vertice v) {
        if (mode==1)                      //falscher Modus
            return;
        Edge e = neighbourtreeID.extract(new Edge(this, v)); //Kante zum Nachbarn suchen und ggf. entfernen
        if (e != null) {               //Wenn die Kante existiert, dann ...
            neighbourtreeW.remove(e); //... auch aus dem gewichtssortierten Baum entfernen und ...
            arrayOK = false;         //... das Array als ungueltig markieren.
        }
    }



    /* testet auf die Existenz einer Kante.
       1. mode==0
        Laufzeit: O(log(n))
       2. mode==1
        Laufzeit: O(n)
     */
    public boolean existsEdge(Vertice v) {
        if (mode == 0)
            return neighbourtreeID.exists(new Edge(this, v));  //Bei einem AVL-Baum geht's schnell.
        else {
            if (!arrayOK) {                                  //Bei einer Liste muessen wir jedoch
                createArray();                              //jedes Element der Liste (bzw. des
                arrayOK = true;                            //Arrays) daraufhin ueberpruefen, ob der
            }                                             //Arrays) daraufhin ueberpruefen, ob der
            for (int i = 0; i < neighbours.length; i++)  //zweite Endknoten der gesuchte ist. Da
                if (neighbours[i].toVertice.id == v.id) //dies ohnehin linearen Aufwand hat, koennen
                    return true;                       //wir auch gleich das Array aktualisieren.
            return false;
        }
    }



    /* gibt die Kante zum Nachbarn v zurueck.
       1. mode==0
        Laufzeit: O(log(n))
       2. mode==1
        Laufzeit: O(n)
     */
    public Edge getEdge(Vertice v) {
        if (mode == 0)
            return neighbourtreeID.valueOf(new Edge(this, v)); //Bei einem AVL-Baum geht' schnell.
        else {
            if (!arrayOK) {                                  //Bei einer Liste muessen wir jedoch
                createArray();                              //jedes Element der Liste (bzw. des
                arrayOK = true;                            //Arrays) daraufhin ueberpruefen, ob der
            }                                             //Arrays) daraufhin ueberpruefen, ob der
            for (int i = 0; i < neighbours.length; i++)  //zweite Endknoten der gesuchte ist. Da
                if (neighbours[i].toVertice.id == v.id) //dies ohnehin linearen Aufwand hat, koennen
                    return neighbours[i];              //wir auch gleich das Array aktualisieren.
            return null;
        }
    }



    /* gibt die Anzahl der Nachbarn zurueck.
       Laufzeit: O(1)
     */
    public int countEdge() {
        if (mode == 0)
            return neighbourtreeID.size();
        else
            return neighbourlist.count();
    }



    /* berechnet das Adjazenz-Array neu.
       Laufzeit: O(n)
     */
    public void createArray() {
        neighbours = new Edge[countEdge()];       //Array erzeugen
        if (mode==0)
            neighbourtreeW.toArray(neighbours); //Array fuellen
        else
            neighbourlist.toArray(neighbours);
        arrayOK = true;                      //Array ist gueltig.
    }



    /* schaltet in den Modus mode um.
       1. mode==0 => mode==0
        Laufzeit: O(1)
       2. mode==0 => mode==1
        Laufzeit: O(n)
       3. mode==1 => mode==0
        Laufzeit: O(n*log(n))
       4. mode==1 => mode==1
        Laufzeit: O(1)
     */
    public void switchMode(int mode) {
        if (mode == this.mode)            //Der Modus stimmt schon.
            return;
        if (mode == 0) {                //Umschaltung: 1 => 0
            neighbourtreeW  = new AVLTree<Edge>(new EdgeWeightComparator());
            neighbourtreeID = new AVLTree<Edge>(new EdgeIDComparator());
            if (!arrayOK)
                createArray();
            mode = 1;
            for (int i = 0; i < neighbours.length; i++)
                addEdge(neighbours[i].toVertice, neighbours[i].weight);
            createArray();
        } else {
            neighbourlist = new List<Edge>();
            if (!arrayOK)
                createArray();
            mode = 0;
            neighbourlist.fromArray(neighbours);
        }
    }



    /* gibt eine Zeichenfolgendarstellung zurueck.
       Laufzeit: O(1)
     */
    public String toString() {
        return "("+id+")";
    }





    /* Ich bin eine Vergleichs-Black-Box fuer Kanten, die zunaechst nach dem Gewicht sortiert.
       Alle Laufzeiten (es gibt ja nur die Vergleichsmethode) sind O(1).
     */
    private class EdgeWeightComparator implements Comparator<Edge> {


        /* vergleicht zwei Kanten miteinander.
           1. Vergleich nach Gewicht
           2. Vergleich nach Maximum der Endknoten-IDs (bei Gleichheit bei 1.)
           3. Vergleich nach Minimum der Endknoten-IDs (bei Gleichheit bei 1. und 2.)
          Laufzeit: O(1)
         */
        public int compare (Edge e1, Edge e2) {
            int i = (int) Math.signum(e1.weight - e2.weight); //Sortierung nach Gewicht, ...
            if (i == 0)
                i = (int) Math.signum( Math.max(e1.fromVertice.id, e1.toVertice.id) // ... Maximum und ...
                                      -Math.max(e2.fromVertice.id, e2.toVertice.id));
            return (i != 0? i : (int) Math.signum(                   //... Minimum der Knoten-IDs.
                       Math.min(e1.toVertice.id, e1.fromVertice.id)
                     - Math.min(e2.toVertice.id, e2.fromVertice.id)));
        }
    }





    /* Ich bin eine Vergleichs-Black-Box fuer Kanten, die nach den ids sortiert.
       Alle Laufzeiten (es gibt ja nur die Vergleichsmethode) sind O(1).
     */
    private class EdgeIDComparator implements Comparator<Edge> {



        /* vergleicht zwei Kanten mit einander.
           1. Vergleich nach Maximum der Endknoten-IDs
           2. Vergleich nach Minimum der Endknoten-IDs (bei Gleichheit bei 1.)
          Laufzeit: O(1)
         */
        public int compare (Edge e1, Edge e2) {
            int i = (int) Math.signum(                               // Vergleich von Maximum und ...
                       Math.max(e1.fromVertice.id, e1.toVertice.id)
                      -Math.max(e2.fromVertice.id, e2.toVertice.id));
            return (i != 0? i : (int) Math.signum(                //... Minimum der Knoten-IDs.
                       Math.min(e1.toVertice.id, e1.fromVertice.id)
                     - Math.min(e2.toVertice.id, e2.fromVertice.id)));
        }
    }
}